import matplotlib.pyplot as plt

# https://stackoverflow.com/a/42972469/1549476


class data_linewidth_plot:
    def __init__(self, x, y, **kwargs):
        self.ax = kwargs.pop("ax", plt.gca())
        self.fig = self.ax.get_figure()
        self.lw_data = kwargs.pop("linewidth", 1)
        self.lw = 1
        self.fig.canvas.draw()

        self.ppd = 72.0 / self.fig.dpi
        self.trans = self.ax.transData.transform
        (self.linehandle,) = self.ax.plot([], [], **kwargs)
        if "label" in kwargs:
            kwargs.pop("label")
        (self.line,) = self.ax.plot(x, y, **kwargs)
        self.line.set_color(self.linehandle.get_color())
        self._resize()
        self.cid = self.fig.canvas.mpl_connect("draw_event", self._resize)

    def _resize(self, event=None):
        lw = ((self.trans((1, self.lw_data)) - self.trans((0, 0))) * self.ppd)[1]
        if lw != self.lw:
            self.line.set_linewidth(lw)
            self.lw = lw
            self._redraw_later()

    def _redraw_later(self):
        self.timer = self.fig.canvas.new_timer(interval=10)
        self.timer.single_shot = True
        self.timer.add_callback(lambda: self.fig.canvas.draw_idle())
        self.timer.start()



from abc import ABC, abstractmethod
from dataclasses import dataclass

from matplotlib import pyplot as plt

from .coord import Coord
from .label import Label


@dataclass(frozen=True)
class Station(ABC):
    coord: Coord
    label: Label

    def with_coord(self, coord):
        return type(self)(coord, self.label)

    @abstractmethod
    def get_markers(self, x, y, r, black, white):
        pass

    def with_label(self, label):
        return type(self)(self.coord, label)

    def with_positioning(self, neighbor):
        return type(self)(
            self.coord,
            self.label.absorb_positioning(neighbor.label),
        )

    def has_positioning(self):
        return self.label.positioning_info() == ["loc", "ang"]

    def positioning_info(self):
        return self.label.positioning_info()


@dataclass(frozen=True)
class BasicStation(Station):
    def get_markers(self, x, y, r, black, white):
        return [
            plt.Circle((x, y), r * 0.5, color=white),
        ]

    @property
    def font_size(self):
        return 1


@dataclass(frozen=True)
class MajorStation(Station):
    def get_markers(self, x, y, r, black, white):
        return [
            plt.Circle((x, y), r * 0.65, color=black),
            plt.Circle((x, y), r * 0.4, color=white),
        ]

    @property
    def font_size(self):
        return 1.5


def make_station(coord, name, loc, ang=0, kind="basic"):
    if kind == "basic":
        typ = BasicStation
    elif kind == "major":
        typ = MajorStation
    else:
        raise ValueError(f"unknown station kind: {kind}")
    return typ(coord, Label(name=name, loc=loc, ang=ang))



from abc import ABC, abstractmethod


class Scheme(ABC):
    @abstractmethod
    def get_basic_color(self, color_name):
        pass

    @abstractmethod
    def rcparams(self):
        pass

    def get_color(self, color_name):
        if color_name == "bg":
            return self.get_basic_color("white")
        if color_name == "fg":
            return self.get_basic_color("black")
        return self.get_basic_color(color_name)

    def modify_label(self, label):
        return label


class MBTA(Scheme):
    def get_basic_color(self, color_name):
        return {
            "red": "#e43037",
            "orange": "#f7941e",
            "yellow": "#fdd808",
            "green": "#009867",
            "blue": "#1b91d1",
            "purple": "#8c4799",
            "pink": "#e27ea6",
            "white": "#ffffff",
            "black": "#000000",
        }[color_name]

    def rcparams(self):
        return dict()

    def modify_label(self, label):
        return label.title()


class Retro(Scheme):
    def get_basic_color(self, color_name):
        return {
            "red": "#9c1f29",
            "orange": "#b62",
            "yellow": "#d8a71f",
            "green": "#395d39",
            "blue": "#3d447d",
            "purple": "#5d395d",
            "pink": "#d8a71f",
            "white": "#e8dacf",
            "black": "#000000",
        }[color_name]

    def rcparams(self):
        return {"font.family": "Futura"}

    def modify_label(self, label):
        return label.upper()


class Terminal(Scheme):
    def get_basic_color(self, color_name):
        return {
            "red": "#ff0000",
            "orange": "#ff8a00",
            "yellow": "#ffff00",
            "green": "#00ff00",
            "blue": "#0000ff",
            "purple": "#ff00ff",
            "pink": "#ff00ff",
            "white": "#000000",
            "black": "#ffffff",
        }[color_name]

    def rcparams(self):
        return {"font.family": "Courier New", "font.weight": "bold"}

    def modify_label(self, label):
        return label.lower()



from matplotlib import pyplot as plt
import numpy as np
from metrodraw.model.coord import Coord
from metrodraw.utils.data_linewidth_plot import data_linewidth_plot


class Renderer:
    def __init__(self, scheme, dpi=200, lw=0.2, font_size=4):
        self.lw = lw
        self.dpi = dpi
        self.ax = None
        self.font_size = font_size
        self._handles = []
        self.scheme = scheme
        self.original_rcparams = None

    def __enter__(self):
        self.original_rcparams = {
            k: plt.rcParams[k] for k in self.scheme.rcparams().keys()
        }
        for k, v in self.scheme.rcparams().items():
            plt.rcParams[k] = v
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        assert self.original_rcparams is not None, "Renderer.__enter__ was not called"
        for k, v in self.original_rcparams.items():
            plt.rcParams[k] = v
        self.original_rcparams = None

    @property
    def coord_config(self):
        return dict(interlining_offset=self.lw * 0.7)

    def line(self, line):
        for seg in line.segments:
            xs, ys = seg.coords(self.coord_config)
            self._handles += [
                data_linewidth_plot(
                    xs,
                    ys,
                    ax=self.ax,
                    linewidth=self.lw,
                    solid_capstyle="round",
                    # solid_joinstyle="miter",
                    color=self.scheme.get_color(line.color),
                )
            ]
            assert len(xs) == 2
            assert len(ys) == 2

            cx = (xs[0] + xs[1]) / 2
            cy = (ys[0] + ys[1]) / 2
            angle = np.rad2deg(np.arctan2(ys[1] - ys[0], xs[1] - xs[0]))

            self.ax.text(
                x=cx,
                y=cy,
                s=seg.label,
                ha="center",
                va="center",
                rotation=angle,
                fontsize=self.lw * 30,
                color=self.scheme.get_color("bg"),
            )

    def station(self, station, no_label=False):
        # draw circle at station
        x, y = station.coord.get_x(self.coord_config), station.coord.get_y(
            self.coord_config
        )
        r = self.lw / 2
        for i, mark in enumerate(
            station.get_markers(
                x,
                y,
                r,
                black=self.scheme.get_color("fg"),
                white=self.scheme.get_color("bg"),
            )
        ):
            # bring to front
            mark.zorder = 10 + i / 10
            self.ax.add_artist(mark)

        if not no_label:
            self.label(x, y, r, station.font_size, station.label)

    def interlining(self, coord, station):
        self.station(station.with_coord(coord), no_label=True)
        x1, y1 = station.coord.get_x(self.coord_config), station.coord.get_y(
            self.coord_config
        )
        x2, y2 = coord.get_x(self.coord_config), coord.get_y(self.coord_config)

        self._handles += [
            data_linewidth_plot(
                [x1, x2],
                [y1, y2],
                ax=self.ax,
                linewidth=self.lw / 4,
                solid_capstyle="round",
                # solid_joinstyle="miter",
                color=self.scheme.get_color("white"),
            )
        ]

    def label(self, x, y, r, font_size, label):
        # draw station name
        loc = label.loc  # one of l, r, u, d, ul, ur, dl, dr

        loc_new = Coord(x, y).move(loc, length=r * 3)

        loc = "".join(sorted(loc))  # normalize

        ha, va = {
            "l": ("right", "center"),
            "r": ("left", "center"),
            "u": ("center", "bottom"),
            "d": ("center", "top"),
            "lu": ("right", "bottom"),
            "ru": ("left", "bottom"),
            "dl": ("right", "top"),
            "dr": ("left", "top"),
        }[loc]

        self.ax.text(
            loc_new.x,
            loc_new.y,
            self.scheme.modify_label(label.name),
            ha=ha,
            va=va,
            rotation=label.ang,
            fontsize=font_size * self.font_size,
            color=self.scheme.get_color("fg"),
        )

    def railmap(self, railmap):
        xlow, xhigh, ylow, yhigh = railmap.bounds(self.coord_config, 1)

        size = xhigh - xlow, yhigh - ylow
        plt.figure(dpi=self.dpi, facecolor=self.scheme.get_color("bg"), figsize=size)
        self.ax = plt.gca()

        for line in railmap.lines:
            self.line(line)
        for station in set(railmap.stations):
            self.station(station)
        for coord, station in railmap.interlining_stations:
            self.interlining(coord, station)

        self.ax.axis("square")
        self.ax.axis("off")
        self.ax.set_xlim(xlow, xhigh)
        self.ax.set_ylim(ylow, yhigh)


def render(railmap, path, **kwargs):
    with Renderer(**kwargs) as renderer:
        renderer.railmap(railmap)
        plt.savefig(path)
        plt.savefig(path)



from dataclasses import dataclass


@dataclass(frozen=True)
class Label:
    name: str
    loc: str
    ang: float

    def absorb_positioning(self, neighbor):
        return Label(
            self.name,
            neighbor.loc if self.loc is None else self.loc,
            neighbor.ang if self.ang is None else self.ang,
        )

    def positioning_info(self):
        return [x for x in ["loc", "ang"] if getattr(self, x) is not None]



from dataclasses import dataclass


@dataclass(frozen=True)
class Coord:
    x: float
    y: float

    def off(self, dx, dy):
        return type(self)(self.x + dx, self.y + dy)

    def move(self, direct, scale=1, length=None):
        if isinstance(direct, Coord):
            return direct
        direct = direct.lower()
        dx = 0
        dy = 0
        dy += direct.count("u")
        dy -= direct.count("d")
        dx += direct.count("r")
        dx -= direct.count("l")
        if length is not None:
            assert scale == 1
            scale = length / (dx ** 2 + dy ** 2) ** 0.5
        dx, dy = dx * scale, dy * scale
        return self.off(dx, dy)

    def get_x(self, config):
        return self.x

    def get_y(self, config):
        return self.y


@dataclass(frozen=True)
class InterliningCoord:
    coord: Coord
    interlining_direction: str

    def off(self, dx, dy):
        return InterliningCoord(self.coord.off(dx, dy), self.interlining_direction)

    def move(self, direct, scale=1, length=None):
        return InterliningCoord(
            self.coord.move(direct, scale=scale, length=length),
            self.interlining_direction,
        )

    def get_x(self, config):
        return self.coord.move(
            self.interlining_direction,
            scale=config["interlining_offset"],
        ).x

    def get_y(self, config):
        return self.coord.move(
            self.interlining_direction,
            scale=config["interlining_offset"],
        ).y



from metrodraw.model.coord import Coord, InterliningCoord
from metrodraw.model.station import make_station


class LineSegment:
    def __init__(self, start, end, label):
        self.start = start
        self.end = end
        self.label = label

    def coords(self, config):
        return [self.start.get_x(config), self.end.get_x(config)], [
            self.start.get_y(config),
            self.end.get_y(config),
        ]


class Line:
    def __init__(self, railmap, start, color, name):
        self.railmap = railmap
        self.last = start
        self.segment_basis = start
        self.color = color
        self.name = name
        self.segments = []
        self.grid_size = 1
        self._interlining = None

    def stations(self, *names, **kwargs):
        return [self.station(name, **kwargs) for name in names]

    def segment(self, direction, label=None):
        coord = self.last
        original = self.segment_basis
        coord = coord.move(direction, scale=self.grid_size)
        self.goto(coord)
        coord = self.with_interlining(coord)
        self.segments.append(LineSegment(original, coord, label))
        self.railmap.add_neighboring(original, coord)
        return coord

    def station(
        self,
        name,
        direction=None,
        *,
        loc=None,
        ang=None,
        kind="basic",
        segment_label=None,
    ):
        if direction is not None:
            self.segment(direction, segment_label)
        station_loc = self.with_interlining(self.last)
        self.railmap.add_station(
            make_station(station_loc, name, loc, ang=ang, kind=kind)
        )
        return station_loc

    def goto(self, coord):
        self.last = coord
        self.segment_basis = self.with_interlining(coord)

    def grid(self, size):
        self.grid_size = size

    def interlining(self, direction, when="next"):
        self._interlining = direction
        assert when in ["now", "next"], f"Invalid when: {when}, must be now or next"
        if when == "now":
            self.segment_basis = self.with_interlining(self.last)

    def no_interlining(self, keep_pos=False):
        self._interlining = None
        if keep_pos:
            self.last = self.segment_basis

    def with_interlining(self, coord):
        if self._interlining is None:
            return coord
        wi = InterliningCoord(coord, self._interlining)
        self.railmap.add_interlining(coord, wi)
        return wi



from metrodraw.model.label import Label
from metrodraw.model.coord import Coord, InterliningCoord
from metrodraw.model.station import make_station
from .line import Line


class Railmap:
    def __init__(self):
        self.lines = []
        self.stations = []
        self.interlining_stations = []
        self.coord_to_station = {}
        self.neighboring_coords = set()

    def line(self, *args):
        self.lines.append(Line(self, *args))
        return self.lines[-1]

    def add_station(self, station):
        lookup_coord = station.coord
        if isinstance(lookup_coord, InterliningCoord):
            lookup_coord = lookup_coord.coord

        if lookup_coord in self.coord_to_station:
            old_station = self.stations[self.coord_to_station[lookup_coord]]
            new_label = station.label
            old_label = old_station.label
            assert (
                new_label.name == old_label.name
            ), f"Attempted to add station {new_label.name} at {lookup_coord} but already have station {old_label.name} at that location"
            merged_label = merge_labels(new_label, old_label)
            self.stations[self.coord_to_station[lookup_coord]] = old_station.with_label(
                merged_label
            )
        else:
            self.stations.append(station)
            self.coord_to_station[station.coord] = len(self.stations) - 1

    def add_neighboring(self, coord_a, coord_b):
        self.neighboring_coords.add((coord_a, coord_b))

    def add_interlining(self, orig, interlined):
        self.add_neighboring(orig, interlined)
        sa = self.station_at(orig)
        if sa is None:
            return
        self.interlining_stations.append((interlined, sa))

    @property
    def neighboring_stations(self):
        coord_to_stations = {}
        for station in self.stations:
            coord_to_stations.setdefault(station.coord, set()).add(station)
        result = {}
        for coord_a, coord_b in self.neighboring_coords:
            for station_a in coord_to_stations.get(
                coord_a, [make_station(coord_a, "dummy", loc=None)]
            ):
                for station_b in coord_to_stations.get(
                    coord_b, [make_station(coord_b, "dummy", loc=None)]
                ):
                    result.setdefault(station_a, set()).add(station_b)
                    result.setdefault(station_b, set()).add(station_a)
        return result

    def propagate_positioning_map(self):
        ns = self.neighboring_stations
        result = {station: station for station in ns}
        if not any(station.has_positioning() for station in result):
            info = sorted(
                {field for station in result for field in station.positioning_info()}
            )
            info_kwargs = {"loc": "l", "ang": 0}
            for field in info:
                info_kwargs[field] = None
            neighbor = make_station(Coord(0, 0), "dummy", **info_kwargs)
            result = {station: station.with_positioning(neighbor) for station in result}
        while True:
            dirty = False
            for station in result:
                dirty = dirty or not result[station].has_positioning()
                for neighbor in ns[station]:
                    if result[station].has_positioning():
                        break
                    result[station] = result[station].with_positioning(result[neighbor])
            if not dirty:
                break
        return result

    def propagate_positioning(self):
        positioning_map = self.propagate_positioning_map()
        self.stations = [positioning_map[station] for station in self.stations]

    def bounds(self, config, pad=0):
        xs = [station.coord.get_x(config) for station in self.stations]
        ys = [station.coord.get_y(config) for station in self.stations]

        xmin, xmax = min(xs), max(xs)
        ymin, ymax = min(ys), max(ys)

        xmin -= pad
        xmax += pad
        ymin -= pad
        ymax += pad

        return xmin, xmax, ymin, ymax

    def station_at(self, coord):
        if coord not in self.coord_to_station:
            return None
        return self.stations[self.coord_to_station[coord]]


def merge_labels(a, b):
    if a.name != b.name:
        raise ValueError(f"Cannot merge labels {a.name} and {b.name}")
    loc = a.loc if a.loc is not None else b.loc
    if a.loc != b.loc:
        assert (
            a.loc is None or b.loc is None
        ), f"Cannot merge labels with different locs: {a.loc} and {b.loc}"

    ang = a.ang if a.ang is not None else b.ang
    if a.ang != b.ang:
        assert (
            a.ang is None or b.ang is None
        ), f"Cannot merge labels with different angs: {a.ang} and {b.ang}"

    return Label(a.name, loc, ang)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("metrodraw/version.py", "r") as fh:
    version = fh.read().strip()

setuptools.setup(
    name="metrodraw",
    version=version,
    author="Kavi Gupta",
    author_email="metrodraw@kavigupta.org",
    description="Allows you to create metro maps.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kavigupta/metrodraw",
    packages=setuptools.find_packages(),
    entry_points={},
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=["matplotlib==3.7.4"],
)
